//@+leo

//@+node:0::@file src/xfun.c
//@+body
//@@language c
/*
    rox-tasklist    provides a tasklist for the rox desktop
    Copyright (C) 2002  Robert Lowe

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author may be contacted at pngwen@acm.org
*/

//@<< xfun #includes >>
//@+node:1::<< xfun #includes >>
//@+body
#include "xfun.h"
#include "task-gui.h"
#include <gdk-pixbuf/gdk-pixbuf.h>

//@-body
//@-node:1::<< xfun #includes >>


//@<< xfun declarations >>
//@+node:2::<< xfun declarations >>
//@+body
gulong XA_COMPOUND_TEXT = 0;
gulong XA_KWM_WIN_ICON = 0;
gulong XA_WIN_STATE = 0;
gulong WIN_CLIENT_LIST=0;

#define WIN_SHADED(a) a & (1<<5)
#define WIN_HIDDEN(a) a & (1<<6)

int xevent_ErrorHandler(Display* dpy, XErrorEvent *event);

//@-body
//@-node:2::<< xfun declarations >>


//@+others
//@+node:3::xfun_init
//@+body
//@+at
//  inits x stuff and returns the default tasklist

//@-at
//@@c

GList* xfun_init()
{
  GdkWindow* window;
  XWindowAttributes attribs={0,};

  taskbutton_height=20;
  taskbutton_width=20;

  taskbutton_count=0;
  XA_COMPOUND_TEXT=gdk_atom_intern("COMPOUND_TEXT", FALSE);
  XA_KWM_WIN_ICON=gdk_atom_intern("KWM_WIN_ICON", FALSE);
  XA_WIN_STATE=gdk_atom_intern("_WIN_STATE", FALSE);
  WIN_CLIENT_LIST=gdk_atom_intern("_WIN_CLIENT_LIST", FALSE);

  /* install our x error handler*/
  XSetErrorHandler(xevent_ErrorHandler);

  window=gdk_window_ref_from_xid(GDK_ROOT_WINDOW()); 
  XGetWindowAttributes (GDK_DISPLAY (),
			GDK_ROOT_WINDOW (),
			&attribs);
  XSelectInput (GDK_DISPLAY (),
		GDK_ROOT_WINDOW (),
		attribs.your_event_mask |
		StructureNotifyMask |
		SubstructureNotifyMask |
		PropertyChangeMask);

  make_task_list(1);
  gdk_window_add_filter(window, root_event_monitor, 0);
  return tasklist;
}

//@-body
//@-node:3::xfun_init
//@+node:4::Functions Adapted from gwmh
//@+node:1::get_typed_property_data
//@+body
gpointer
get_typed_property_data (Display *xdisplay,
			 Window   xwindow,
			 Atom     property,
			 Atom     requested_type,
			 gint    *size_p,
			 guint    expected_format)
{
  static const guint prop_buffer_lengh = 1024 * 1024;
  unsigned char *prop_data = NULL;
  Atom type_returned = 0;
  unsigned long nitems_return = 0, bytes_after_return = 0;
  int format_returned = 0;
  gpointer data = NULL;
  gboolean abort = FALSE;

  g_return_val_if_fail (size_p != NULL, NULL);
  *size_p = 0;

  gdk_error_trap_push ();

  abort = XGetWindowProperty (xdisplay,
			      xwindow,
			      property,
			      0, prop_buffer_lengh,
			      False,
			      requested_type,
			      &type_returned, &format_returned,
			      &nitems_return,
			      &bytes_after_return,
			      &prop_data) != Success;
  if (gdk_error_trap_pop () ||
      type_returned == None)
    abort++;
  if (!abort &&
      requested_type != AnyPropertyType &&
      requested_type != type_returned)
    {
      /* aparently this can happen for some properties of broken apps, be silent */
      abort++;
    }
  if (!abort && bytes_after_return)
    {
      g_warning (G_GNUC_PRETTY_FUNCTION "(): Eeek, property has more than %u bytes, stored on harddisk?",
		 prop_buffer_lengh);
      abort++;
    }
  if (!abort && expected_format && expected_format != format_returned)
    {
      g_warning (G_GNUC_PRETTY_FUNCTION "(): Expected format (%u) unmatched (%d), programmer was drunk?",
		 expected_format, format_returned);
      abort++;
    }
  if (!abort && prop_data && nitems_return && format_returned)
    {
      switch (format_returned)
	{
	case 32:
	  *size_p = nitems_return * 4;
	  if (sizeof (gulong) == 8)
	    {
	      guint32 i, *mem = g_malloc0 (*size_p + 1);
	      gulong *prop_longs = (gulong*) prop_data;

	      for (i = 0; i < *size_p / 4; i++)
		mem[i] = prop_longs[i];
	      data = mem;
	    }
	  break;
	case 16:
	  *size_p = nitems_return * 2;
	  break;
	case 8:
	  *size_p = nitems_return;
	  break;
	default:
	  g_warning ("Unknown property data format with %d bits (extraterrestrial?)",
		     format_returned);
	  break;
	}
      if (!data && *size_p)
	{
	  guint8 *mem = NULL;

	  if (format_returned == 8 && type_returned == XA_COMPOUND_TEXT)
	    {
	      gchar **tlist = NULL;
	      gint count = gdk_text_property_to_text_list (type_returned, 8, prop_data,
							   nitems_return, &tlist);

	      if (count && tlist && tlist[0])
		{
		  mem = (guint8 *)g_strdup (tlist[0]);
		  *size_p = strlen ((char *)mem);
		}
	      if (tlist)
		gdk_free_text_list (tlist);
	    }
	  if (!mem)
	    {
	      mem = g_malloc (*size_p + 1);
	      memcpy (mem, prop_data, *size_p);
	      mem[*size_p] = 0;
	    }
	  data = mem;
	}
    }

  if (prop_data)
    XFree (prop_data);
  
  return data;
}

//@-body
//@-node:1::get_typed_property_data
//@+node:2::check_client
//@+body
//@+at
//  checks to see if the client is an actual xwindow client. Also grabs the  state of the window. for the task

//@-at
//@@c

gboolean
check_client (Atom state_atom, task_type* task)
{
  gboolean valid_client = TRUE;

  if (valid_client)
    {
      Atom dummy1;
      int format = 0;
      unsigned long dummy2, nitems = 0, *prop = NULL;
      
      XGetWindowProperty (task->display, task->xwindow, state_atom, 0, 1024, 
                          False, state_atom, &dummy1, &format, 
                          &nitems, &dummy2, (unsigned char **) &prop);
      if (prop)
	{
	  valid_client = format == 32 && 
                            nitems > 0 &&
                            prop[0] == NormalState;
          if(valid_client)
            task->state=NORMAL;
          else if(prop[0]==IconicState)
	  {
	    task->state=ICONIC;
	    return TRUE;
	  }
	  else if(prop[0]==InactiveState)
	  {
	    task->state=NORMAL;
	    return TRUE;
	  }
	  XFree (prop);
	}
      else
        valid_client=0;
    }

  if (valid_client)
    {
      XWindowAttributes attributes = { 0, };

      XGetWindowAttributes (task->display, task->xwindow, &attributes);
      valid_client = (attributes.class == InputOutput); 
    }

  if (valid_client)
    {
      XWMHints *hints = XGetWMHints (task->display, task->xwindow);
      
      valid_client &= hints && (hints->flags & InputHint) && hints->input;
      if (hints)
	XFree (hints);
    }
  
  return valid_client;
}

//@-body
//@-node:2::check_client
//@+node:3::find_input_client
//@+body
//@+at
//  finds the client and sets the window in the task

//@-at
//@@c

void find_input_client(Atom state_atom, task_type* task)
{
  Window dummy1, dummy2, *children = NULL;
  unsigned int n_children = 0;
  guint i;
  
  if (check_client (state_atom, task))
    return;
  
  if (!XQueryTree (task->display, task->xwindow, 
                   &dummy1, &dummy2, &children, &n_children) || !children)
  {
    task->xwindow=None;
    return;
  }
  
  for (i = 0; i < n_children; i++)
    {
      task->xwindow=children[i];
      find_input_client (state_atom, task);
      if (task->xwindow)
	break;
    }
  
  XFree (children);
 
  return;
}

//@-body
//@-node:3::find_input_client
//@+node:4::gdk_window_ref_from_xid
//@+body
GdkWindow*
gdk_window_ref_from_xid (Window xwin)
{
  GdkWindow *window;

  /* the xid maybe invalid already, in that case we return NULL */
  window = gdk_window_lookup (xwin);
  if (!window)
    window = gdk_window_foreign_new (xwin);
  else
    gdk_window_ref (window);

  return window;
}

//@-body
//@-node:4::gdk_window_ref_from_xid
//@-node:4::Functions Adapted from gwmh
//@+node:5::Tasklist Functions
//@+node:1::task_compare
//@+body
//@+at
//  a gcompare that compares elem to user. like strcmp

//@-at
//@@c

gint task_compare(gconstpointer elem, gconstpointer user)
{
  task_type* task1;
  task_type* task2;

  if(!elem || !user)
    return -1;

  task1=(task_type*)elem;
  task2=(task_type*)user;
 
  if(task1->xwindow == task2->xwindow)
    return 0;

  if(task1->xwindow < task2->xwindow)
    return -1;

  if(task1->xwindow > task2->xwindow)
    return 1;

  return 1;
}

//@-body
//@-node:1::task_compare
//@+node:2::update_task
//@+body
//@+at
//  update/get the task's info

//@-at
//@@c

void update_task(task_type* task)
{
  Atom dummy1;
  Atom state_atom;
  Window focus;
  unsigned long nitems_return, bytes_after_return;
  int format=0, revert_to;
  unsigned long *prop=NULL;
  gint size=0;
  guint32* flags;
  guint32  bits;
  XWindowAttributes attr;


  /* get the name */
  if(task->name)
    free(task->name);
  XGetWindowProperty(task->display, task->xwindow, XA_WM_NAME, 0, 1024,
                     False, XA_STRING, &dummy1, &format, &nitems_return,
                     &bytes_after_return, &task->name);
			
  /* give nameless tasks a name consisting of one space */		 
  if(!task->name)
  {
    task->name=malloc(2);
	 task->name[0]=' ';
	 task->name[1]='\0';
  }


  /* get the state */
  state_atom=XInternAtom(task->display, "WM_STATE", FALSE);
  XGetWindowProperty (task->display, task->xwindow, state_atom, 0, 1024, 
                      False, state_atom, &dummy1, &format, &nitems_return,
                      &bytes_after_return, (unsigned char **) &prop);
  if(prop&& prop[0]==IconicState)
  {
    task->state=ICONIC;
    /* first time around this is set elsewhere */
    if(task->button)
      taskbutton_iconify(task->button);
  }
  else
  {
    task->state=NORMAL;
    /* first time around this is set elsewhere */
    if(task->button)
      taskbutton_normalfy(task->button);
  }
  free(prop); 
  /* check focus */
  XGetInputFocus(task->display, &focus, &revert_to); 
  if(task->xwindow==focus)
    task->state=FOCUSED;

  /* check shaded */
  flags=get_typed_property_data(task->display, task->xwindow,
			       XA_WIN_STATE,
			       XA_CARDINAL,
			       &size, 32);

  bits=flags? flags[0] : 0;
  if(WIN_SHADED(bits))
    task->state=SHADED;
  
 
  XGetWindowAttributes(task->display, task->xwindow, &attr);
  /* update display info */
  if((attr.map_state==IsUnmapped && task->state!=ICONIC && task->state!=SHADED) ||
     ( flags && WIN_HIDDEN(bits)))
  {
    if(task->button)
      taskbutton_destroy(task->button);
    task->button=0;
  }
  else
  {
    if(!task->button)
    {
      task->button=taskbutton_new(task);
      taskbutton_display(task->button);
    }
    update_task_display(task->button);
  }
  g_free(flags);

}

//@-body
//@-node:2::update_task
//@+node:3::add_task
//@+body
//@+at
//  gets all the information for a task. The xwindow field of task must be set.  This will find the client window and retrieve all 
// the needed info   it also sets up the filters for the task and adds it to the task list

//@-at
//@@c

void add_task(task_type* task) 
{
  XWindowAttributes attribs = { 0, };
  GList* node;

  find_input_client(XInternAtom(task->display, "WM_STATE", FALSE), task);
  if(task->xwindow)
  {
    if((node=g_list_find_custom(tasklist, task, task_compare)))
    {
      ((task_type*) node->data)->visited=1;
      task->xwindow=0;
      return;
    }
    task->button=0;
    task->name=0;
    task->gdkframe=0;
    /* get task's info */ 
    update_task(task);
    if(!task->xwindow)
    {
      return;
    }
    /* see if we should ignore this task */
	 if(ignore_list_check(task->name))
	 {
	   if(task->button)
	     taskbutton_destroy(task->button);
	   free(task->name);
	   task->xwindow=0;
		return;
	 }

    task->visited=1;
    tasklist=g_list_append(tasklist, task);

    /* set up filtering */
    task->gdkframe=gdk_window_ref_from_xid(task->xwindow);
    gdk_window_add_filter(task->gdkframe, task_event_monitor, 
                          task);
    XGetWindowAttributes(task->display, task->xwindow, &attribs);
    XSelectInput(task->display, task->xwindow, attribs.your_event_mask |
                   StructureNotifyMask|PropertyChangeMask|FocusChangeMask|
                   SubstructureNotifyMask); 

    /* handle gui stuff */
    if(task->button)
    {
      switch (task->state)
      {
        case NORMAL:  taskbutton_normalfy(task->button);
                      break;
        case ICONIC:  taskbutton_iconify(task->button);
	              break;
        case FOCUSED: taskbutton_focus(task->button);
	              break;
        default: break;
      }
    }
  }
}

//@-body
//@-node:3::add_task
//@+node:4::remove_task
//@+body
//@+at
//  remove's a task from the tasklist

//@-at
//@@c

void remove_task(task_type* task)
{
  if(task->gdkframe)
    gdk_window_remove_filter(task->gdkframe, task_event_monitor, task);
  if(task->button)
    taskbutton_destroy(task->button);
  tasklist=g_list_remove(tasklist, task);
}

//@-body
//@-node:4::remove_task
//@+node:5::make_task_list
//@+body
//@+at
//  builds the initial tasklist

//@-at
//@@c

void make_task_list(int mode)
{
  task_type* task=0;
  Window dummy;
  Window* children;
  unsigned int nc;
  static unsigned int oldnc=0;
  int i;
  GList* cur;
  GList* cur2;
  Display* display;

  task=malloc(sizeof(task_type));
  task->display=GDK_DISPLAY(); 
  display=task->display;
  reset_visited();

  
  /* get the first level of the window tree */
  XGrabServer(display);


  XQueryTree(display, GDK_ROOT_WINDOW(), &dummy, &dummy, &children, &nc);
  if(nc==oldnc && !mode)
  {
    XFree(children);
    XUngrabServer(display);
    free(task);
    return;
  }
  oldnc=nc;
  

  for(i=0; i<nc; i++)
  {
    if(!task)
    {
      task=malloc(sizeof(task_type));
      task->display=display; 
    }
    /* find the io client for this child */
    task->xwindow=children[i];
    add_task(task);

    /* if the window was found, we'll need to allocate a new task*/
    if(task->xwindow)
      task=0;
  }

  /* discard task that had no window */
  if(task && !task->xwindow)
  {
    free(task);
	 task=0;
  }

  XUngrabServer(display);
  XFree(children);

  /* purge windows not visited */
  cur=tasklist;
  while(cur)
  {
    /* just in case we delete this node, we'll need this info */
    cur2=cur->next;
    if(!((task_type*)cur->data)->visited)
      remove_task((task_type*)cur->data);
    cur=cur2;
  }
}

//@-body
//@-node:5::make_task_list
//@+node:6::reset_visited
//@+body
void reset_visited()
{
  GList* cur;

  cur=tasklist;

  while(cur)
  {
    ((task_type*)cur->data)->visited=0;
    cur=cur->next;
  }
}

//@-body
//@-node:6::reset_visited
//@+node:7::task_get_mini_icon
//@+body
void
task_get_mini_icon (task_type *task,
		    GdkPixmap **pixmap,
		    GdkBitmap **mask, int* w, int* h)
{
  GdkGC *gc;
  GdkGCPrivate *gc_private;
  GdkPixmapPrivate *private;
  guint32 *atomdata;
  Window xwindow = task->xwindow;
  Window root;
  int x, y;
  guint b, width, height, depth;
  gint size;
  Display *xdisplay;

  g_return_if_fail (task != NULL);
  g_return_if_fail (pixmap != NULL);
  g_return_if_fail (mask != NULL);

  xdisplay = task->display;

  atomdata = get_typed_property_data (xdisplay,
				      xwindow,
				      XA_KWM_WIN_ICON,
				      XA_KWM_WIN_ICON,
				      &size,
				      32);
  if (!atomdata || !atomdata[0] || !size)
    {
      *pixmap = NULL;
      *mask = NULL;

      return;
    }

  /* Get icon size and depth */
  XGetGeometry (xdisplay, (Drawable) atomdata[0], &root, &x, &y, 
		&width, &height, &b, &depth);
  
  /* Create a new GdkPixmap and copy the mini icon pixmap to it */
  *pixmap = gdk_pixmap_new (NULL, width, height, depth);
  gc = gdk_gc_new (*pixmap);
  gc_private = (GdkGCPrivate*) gc;
  private = (GdkPixmapPrivate*) *pixmap;
  XCopyArea (private->xdisplay, atomdata[0], private->xwindow, gc_private->xgc,
	     0, 0, width, height, 0, 0);
  gdk_gc_destroy (gc);
  
  /* get mask size and depth */
  XGetGeometry (xdisplay, (Drawable)atomdata[1], &root, &x, &y, 
		&width, &height, &b, &depth);

  /* Create a new GdkBitmap and copy the mini icon mask to it */
  *mask = gdk_pixmap_new (NULL, width, height, 1);
  gc = gdk_gc_new (*mask);
  gc_private = (GdkGCPrivate*) gc;
  private = (GdkPixmapPrivate*) *mask;
  XCopyArea (private->xdisplay, atomdata[1], private->xwindow, gc_private->xgc,
	     0, 0, width, height, 0, 0);
  gdk_gc_destroy (gc);
  g_free (atomdata);

  *w=width;
  *h=height;
}

//@-body
//@-node:7::task_get_mini_icon
//@-node:5::Tasklist Functions
//@+node:6::Monitors and Handlers
//@+node:1::xevent_ErrorHandler
//@+body



/* this is the error handler for xerrors.  Ignores all badwindow errors, but 
   prints diagnostics on the rest */
int xevent_ErrorHandler(Display* dpy, XErrorEvent *event)
{
  char buff[256];
 
  if(event->error_code != BadWindow)
  {
    XGetErrorText(dpy, event->error_code, buff, sizeof(buff));
    fprintf(stderr, "\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "************************************************\n");
    fprintf(stderr, "Error %d for resource 0x%lx\n", 
	    event->error_code,
	    event->resourceid);
    fprintf(stderr, "************************************************\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\n");
    exit(0);
  }
  return 0;
}

//@-body
//@-node:1::xevent_ErrorHandler
//@+node:2::task_event_monitor
//@+body
GdkFilterReturn 
task_event_monitor(GdkXEvent *gdk_xevent,
                   GdkEvent *event,
                   gpointer taskptr)
{
  task_type* task;
  XEvent *xevent = gdk_xevent;
  int mode=0;

  task=taskptr;
  switch (xevent->type)
  {
    case ConfigureNotify:
      break;
    case PropertyNotify:
      update_task(task);
      break;
    case ReparentNotify:
      break;
    case DestroyNotify:
      mode=1;
    case MapNotify:
      make_task_list(mode);
      break;
    case UnmapNotify:
      break;
    case CirculateNotify:
      break;
    case FocusOut:
      task->state=NORMAL;
      if(task->button)
        taskbutton_unfocus(task->button);
      break;
    case FocusIn:
      task->state=FOCUSED;
      if(task->button)
        taskbutton_focus(task->button);
      break;
    default:
      break;
  }
  return GDK_FILTER_CONTINUE;

}

//@-body
//@-node:2::task_event_monitor
//@+node:3::root_event_monitor
//@+body
GdkFilterReturn 
root_event_monitor (GdkXEvent *gdk_xevent,
		    GdkEvent  *event,
		    gpointer   taskptr)
{
  XEvent *xevent = gdk_xevent;
  int mode=0;
  
  switch (xevent->type)
    {
    case MapNotify:
      mode=1;
    case UnmapNotify:
    case CirculateNotify:
    case ReparentNotify:
    case DestroyNotify:
    case PropertyNotify:
      /*     printf("%s\n", XGetAtomName(GDK_DISPLAY(), xevent->xproperty.atom));
        if(xevent->xproperty.atom==ViewPortAtom)
	printf("\n  Potentially new window %d\n", xevent->xproperty.window); */
      make_task_list(mode);
      break;
    default:
      break;
    }
  
  return GDK_FILTER_CONTINUE;
}

//@-body
//@-node:3::root_event_monitor
//@-node:6::Monitors and Handlers
//@-others


//@-body
//@-node:0::@file src/xfun.c
//@-leo
